@@ -33,7 +33,7 @@ DOMAIN = 'firewalld'
 import gettext
 gettext.install(domain=DOMAIN)
 
-from . import dbus
+from . import dbus # noqa: F401
 
 # configuration
 DAEMON_NAME = 'firewalld'
@@ -42,13 +42,13 @@ APPLET_NAME = 'firewall-applet'
 DATADIR = '/usr/share/' + DAEMON_NAME
 CONFIG_GLADE_NAME = CONFIG_NAME + '.glade'
 COPYRIGHT = '(C) 2010-2017 Red Hat, Inc.'
-VERSION = '0.5.0'
+VERSION = '0.6.0'
 AUTHORS = [
     "Thomas Woerner <twoerner@redhat.com>",
     "Jiri Popelka <jpopelka@redhat.com>",
     "Eric Garver <e@erig.me>",
     ]
-LICENSE = _(
+LICENSE = gettext.gettext(
     "This program is free software; you can redistribute it and/or modify "
     "it under the terms of the GNU General Public License as published by "
     "the Free Software Foundation; either version 2 of the License, or "
@@ -97,7 +97,7 @@ FIREWALLD_PIDFILE = "/var/run/firewalld.pid"
 FIREWALLD_TEMPDIR = '/run/firewalld'
 
 SYSCONFIGDIR = '/etc/conf.d'
-IFCFGDIR = "/etc/conf.d/network-scripts"
+IFCFGDIR = "/etc/sysconfig/network-scripts"
 
 SYSCTL_CONFIG = '/etc/sysctl.conf'
 
@@ -113,10 +113,12 @@ COMMANDS = {
     "modinfo":      "/usr/bin/modinfo",
     "modprobe":     "/usr/bin/modprobe",
     "rmmod":        "/usr/bin/rmmod",
+    "nft":          "/usr/bin/nft",
 }
 
 LOG_DENIED_VALUES = [ "all", "unicast", "broadcast", "multicast", "off" ]
 AUTOMATIC_HELPERS_VALUES = [ "yes", "no", "system" ]
+FIREWALL_BACKEND_VALUES = [ "nftables", "iptables" ]
 
 # fallbacks: will be overloaded by firewalld.conf
 FALLBACK_ZONE = "public"
@@ -127,3 +129,4 @@ FALLBACK_IPV6_RPFILTER = True
 FALLBACK_INDIVIDUAL_CALLS = False
 FALLBACK_LOG_DENIED = "off"
 FALLBACK_AUTOMATIC_HELPERS = "system"
+FALLBACK_FIREWALL_BACKEND = "nftables"
