@@ -26,6 +26,8 @@ from firewall.core.prog import runProg
 from firewall.core.logger import log
 from firewall.functions import tempFile, readfile, splitArgs
 from firewall.config import COMMANDS
+from firewall.core import ipXtables # some common stuff lives there
+from firewall.errors import FirewallError, INVALID_IPV
 import string
 
 PROC_IPxTABLE_NAMES = {
@@ -51,6 +53,8 @@ for table in BUILT_IN_CHAINS.keys():
 
 class ebtables(object):
     ipv = "eb"
+    name = "ebtables"
+    zones_supported = False # ebtables only supported with direct interface
 
     def __init__(self):
         self._command = COMMANDS[self.ipv]
@@ -58,6 +62,7 @@ class ebtables(object):
         self.restore_noflush_option = self._detect_restore_noflush_option()
         self.concurrent_option = self._detect_concurrent_option()
         self.fill_exists()
+        self.available_tables = []
 
     def fill_exists(self):
         self.command_exists = os.path.exists(self._command)
@@ -96,19 +101,45 @@ class ebtables(object):
                                                      " ".join(args), ret))
         return ret
 
-    def _rule_contains(self, rule, pattern):
-        try:
-            i = rule.index(pattern)
-        except ValueError:
-            return False
-        return True
-
     def _rule_validate(self, rule):
         for str in ["%%REJECT%%", "%%ICMP%%", "%%LOGTYPE%%"]:
-            if self._rule_contains(rule, str):
-                raise FirewallError(errors.INVALID_IPV,
+            if str in rule:
+                raise FirewallError(INVALID_IPV,
                         "'%s' invalid for ebtables" % str)
 
+    def is_chain_builtin(self, ipv, table, chain):
+        return table in BUILT_IN_CHAINS and \
+               chain in BUILT_IN_CHAINS[table]
+
+    def build_chain(self, add, table, chain):
+        rule = [ "-t", table ]
+        if add:
+            rule.append("-N")
+        else:
+            rule.append("-X")
+        rule.append(chain)
+        if add:
+            rule += [ "-P", "RETURN" ]
+        return rule
+
+    def build_rule(self, add, table, chain, index, args):
+        rule = [ "-t", table ]
+        if add:
+            rule += [ "-I", chain, str(index) ]
+        else:
+            rule += [ "-D", chain ]
+        rule += args
+        return rule
+
+    def reverse_rule(self, args):
+        return ipXtables.common_reverse_rule(args)
+
+    def check_passthrough(self, args):
+        ipXtables.common_check_passthrough(args)
+
+    def reverse_passthrough(self, args):
+        return ipXtables.common_reverse_passthrough(args)
+
     def set_rules(self, rules, flush=False, log_denied="off"):
         temp_file = tempFile()
 
@@ -179,77 +210,69 @@ class ebtables(object):
         self._rule_validate(rule)
         return self.__run(rule)
 
-    def append_rule(self, rule):
-        self.__run([ "-A" ] + rule)
-
-    def delete_rule(self, rule):
-        self.__run([ "-D" ] + rule)
-
-    def available_tables(self, table=None):
+    def get_available_tables(self, table=None):
         ret = []
         tables = [ table ] if table else BUILT_IN_CHAINS.keys()
         for table in tables:
-            try:
-                self.__run(["-t", table, "-L"])
+            if table in self.available_tables:
                 ret.append(table)
-            except ValueError:
-                log.debug1("ebtables table '%s' does not exist." % table)
+            else:
+                try:
+                    self.__run(["-t", table, "-L"])
+                    self.available_tables.append(table)
+                    ret.append(table)
+                except ValueError:
+                    log.debug1("ebtables table '%s' does not exist." % table)
 
         return ret
 
+    def get_zone_table_chains(self, table):
+        return {}
+
     def used_tables(self):
         return list(BUILT_IN_CHAINS.keys())
 
-    def flush(self, transaction=None):
-        tables = self.used_tables()
-        for table in tables:
+    def build_flush_rules(self):
+        rules = []
+        for table in self.used_tables():
             # Flush firewall rules: -F
             # Delete firewall chains: -X
             # Set counter to zero: -Z
-            msgs = {
-                "-F": "flush",
-                "-X": "delete chains",
-                "-Z": "zero counters",
-            }
             for flag in [ "-F", "-X", "-Z" ]:
-                if transaction is not None:
-                    transaction.add_rule(self.ipv, [ "-t", table, flag ])
-                else:
-                    try:
-                        self.__run([ "-t", table, flag ])
-                    except Exception as msg:
-                        log.error("Failed to %s %s: %s",
-                                  msgs[flag], self.ipv, msg)
+                rules.append(["-t", table, flag])
+        return rules
 
-    def set_policy(self, policy, which="used", transaction=None):
+    def build_set_policy_rules(self, policy, which="used"):
         if which == "used":
             tables = self.used_tables()
         else:
             tables = list(BUILT_IN_CHAINS.keys())
 
+        rules = []
         for table in tables:
             for chain in BUILT_IN_CHAINS[table]:
-                if transaction is not None:
-                    transaction.add_rule(self.ipv,
-                                         [ "-t", table, "-P", chain, policy ])
-                else:
-                    try:
-                        self.__run([ "-t", table, "-P", chain, policy ])
-                    except Exception as msg:
-                        log.error("Failed to set policy for %s: %s", self.ipv,
-                                  msg)
+                rules.append(["-t", table, "-P", chain, policy])
+        return rules
 
-    def apply_default_rules(self, transaction, log_denied="off"):
+    def build_default_tables(self):
+        # nothing to do, they always exist
+        return []
+
+    def build_default_rules(self, log_denied="off"):
+        default_rules = []
         for table in DEFAULT_RULES:
-            if table not in self.available_tables():
+            if table not in self.get_available_tables():
                 continue
-            default_rules = DEFAULT_RULES[table][:]
+            _default_rules = DEFAULT_RULES[table][:]
             if log_denied != "off" and table in LOG_RULES:
-                default_rules.extend(LOG_RULES[table])
+                _default_rules.extend(LOG_RULES[table])
             prefix = [ "-t", table ]
-            for rule in DEFAULT_RULES[table]:
+            for rule in _default_rules:
                 if type(rule) == list:
-                    _rule = prefix + rule
+                    default_rules.append(prefix + rule)
                 else:
-                    _rule = prefix + splitArgs(rule)
-                transaction.add_rule(self.ipv, _rule)
+                    default_rules.append(prefix + splitArgs(rule))
+        return default_rules
+
+    def is_ipv_supported(self, ipv):
+        return ipv == self.ipv
