@@ -21,19 +21,22 @@
 
 __all__ = [ "Firewall" ]
 
-import os.path, sys
+import os.path
+import sys
 import copy
 import time
+import traceback
 from firewall import config
 from firewall import functions
 from firewall.core import ipXtables
 from firewall.core import ebtables
+from firewall.core import nftables
 from firewall.core import ipset
 from firewall.core import modules
 from firewall.core.fw_icmptype import FirewallIcmpType
 from firewall.core.fw_service import FirewallService
-from firewall.core.fw_zone import FirewallZoneIPTables
-from firewall.core.fw_direct import FirewallDirectIPTables
+from firewall.core.fw_zone import FirewallZone
+from firewall.core.fw_direct import FirewallDirect
 from firewall.core.fw_config import FirewallConfig
 from firewall.core.fw_policies import FirewallPolicies
 from firewall.core.fw_ipset import FirewallIPSet
@@ -60,10 +63,10 @@ class Firewall(object):
     def __init__(self):
         self._firewalld_conf = firewalld_conf(config.FIREWALLD_CONF)
 
-        self.ip4tables_backend = ipXtables.ip4tables()
+        self.ip4tables_backend = ipXtables.ip4tables(self)
         self.ip4tables_enabled = True
         self.ip4tables_supported_icmp_types = [ ]
-        self.ip6tables_backend = ipXtables.ip6tables()
+        self.ip6tables_backend = ipXtables.ip6tables(self)
         self.ip6tables_enabled = True
         self.ip6tables_supported_icmp_types = [ ]
         self.ebtables_backend = ebtables.ebtables()
@@ -71,18 +74,15 @@ class Firewall(object):
         self.ipset_backend = ipset.ipset()
         self.ipset_enabled = True
         self.ipset_supported_types = [ ]
-
-        self.available_tables = { }
-        self.available_tables["ipv4"] = self.ip4tables_backend.available_tables()
-        self.available_tables["ipv6"] = self.ip6tables_backend.available_tables()
-        self.available_tables["eb"] = self.ebtables_backend.available_tables()
+        self.nftables_backend = nftables.nftables(self)
+        self.nftables_enabled = True
 
         self.modules_backend = modules.modules()
 
         self.icmptype = FirewallIcmpType(self)
         self.service = FirewallService(self)
-        self.zone = FirewallZoneIPTables(self)
-        self.direct = FirewallDirectIPTables(self)
+        self.zone = FirewallZone(self)
+        self.direct = FirewallDirect(self)
         self.config = FirewallConfig(self)
         self.policies = FirewallPolicies()
         self.ipset = FirewallIPSet(self)
@@ -112,6 +112,7 @@ class Firewall(object):
         self._individual_calls = config.FALLBACK_INDIVIDUAL_CALLS
         self._log_denied = config.FALLBACK_LOG_DENIED
         self._automatic_helpers = config.FALLBACK_AUTOMATIC_HELPERS
+        self._firewall_backend = config.FALLBACK_FIREWALL_BACKEND
         self.nf_conntrack_helper_setting = 0
         self.nf_conntrack_helpers = { }
         self.nf_nat_helpers = { }
@@ -122,28 +123,29 @@ class Firewall(object):
     def _check_tables(self):
         # check if iptables, ip6tables and ebtables are usable, else disable
         if self.ip4tables_enabled and \
-           "filter" not in self.get_available_tables("ipv4"):
+           "filter" not in self.get_backend_by_ipv("ipv4").get_available_tables():
             log.warning("iptables not usable, disabling IPv4 firewall.")
             self.ip4tables_enabled = False
 
         if self.ip6tables_enabled and \
-           "filter" not in self.get_available_tables("ipv6"):
+           "filter" not in self.get_backend_by_ipv("ipv6").get_available_tables():
             log.warning("ip6tables not usable, disabling IPv6 firewall.")
             self.ip6tables_enabled = False
 
         if self.ebtables_enabled and \
-           "filter" not in self.get_available_tables("eb"):
+           "filter" not in self.get_backend_by_ipv("eb").get_available_tables():
             log.warning("ebtables not usable, disabling ethernet bridge firewall.")
             self.ebtables_enabled = False
 
         # is there at least support for ipv4 or ipv6
-        if not self.ip4tables_enabled and not self.ip6tables_enabled:
+        if not self.ip4tables_enabled and not self.ip6tables_enabled \
+           and not self.nftables_enabled:
             log.fatal("No IPv4 and IPv6 firewall.")
             sys.exit(1)
 
     def _start_check(self):
         try:
-            self.ipset_backend.list()
+            self.ipset_backend.set_list()
         except ValueError:
             log.warning("ipset not usable, disabling ipset usage in firewall.")
             # ipset is not usable, no supported types
@@ -151,7 +153,7 @@ class Firewall(object):
             self.ipset_supported_types = [ ]
         else:
             # ipset is usable, get all supported types
-            self.ipset_supported_types = self.ipset_backend.supported_types()
+            self.ipset_supported_types = self.ipset_backend.set_supported_types()
 
         self.ip4tables_backend.fill_exists()
         if not self.ip4tables_backend.restore_command_exists:
@@ -240,6 +242,8 @@ class Firewall(object):
                 value = self._firewalld_conf.get("CleanupOnExit")
                 if value is not None and value.lower() in [ "no", "false" ]:
                     self.cleanup_on_exit = False
+                log.debug1("CleanupOnExit is set to '%s'",
+                           self.cleanup_on_exit)
 
             if self._firewalld_conf.get("Lockdown"):
                 value = self._firewalld_conf.get("Lockdown")
@@ -289,8 +293,15 @@ class Firewall(object):
                     log.debug1("AutomaticHelpers is set to '%s'",
                                self._automatic_helpers)
 
+            if self._firewalld_conf.get("FirewallBackend"):
+                self._firewall_backend = self._firewalld_conf.get("FirewallBackend")
+                log.debug1("FirewallBackend is set to '%s'",
+                           self._firewall_backend)
+
         self.config.set_firewalld_conf(copy.deepcopy(self._firewalld_conf))
 
+        self._select_firewall_backend(self._firewall_backend)
+
         self._start_check()
 
         # load lockdown whitelist
@@ -370,12 +381,19 @@ class Firewall(object):
             try:
                 obj.read()
             except Exception as msg:
-                log.debug1("Failed to load direct rules file '%s': %s",
-                           config.FIREWALLD_DIRECT, msg)
+                log.error("Failed to load direct rules file '%s': %s",
+                          config.FIREWALLD_DIRECT, msg)
         self.direct.set_permanent_config(obj)
         self.config.set_direct(copy.deepcopy(obj))
 
         # automatic helpers
+        #
+        # NOTE: must force loading of nf_conntrack to make sure the values are
+        # available in /proc
+        module_return = self.handle_modules(["nf_conntrack"], True)
+        if module_return:
+            log.error("Failed to load nf_conntrack module: %s" % module_return[1])
+            sys.exit(1)
         if self._automatic_helpers != "system":
             functions.set_nf_conntrack_helper_setting(self._automatic_helpers == "yes")
         self.nf_conntrack_helper_setting = \
@@ -413,6 +431,10 @@ class Firewall(object):
             log.debug1("Unloading firewall modules")
             self.modules_backend.unload_firewall_modules()
 
+        self.apply_default_tables(use_transaction=transaction)
+        transaction.execute(True)
+        transaction.clear()
+
         # apply settings for loaded ipsets while reloading here
         if self.ipset_enabled and self.ipset.has_ipsets():
             log.debug1("Applying ipsets")
@@ -603,11 +625,6 @@ class Firewall(object):
                 self.config.forget_zone(combined_zone.name)
             self.zone.add_zone(combined_zone)
 
-    def get_available_tables(self, ipv):
-        if ipv in [ "ipv4", "ipv6", "eb" ]:
-            return self.available_tables[ipv]
-        return [ ]
-
     def cleanup(self):
         self.icmptype.cleanup()
         self.service.cleanup()
@@ -667,15 +684,99 @@ class Firewall(object):
                         del self._module_refcount[module]
         return None
 
-    # apply default rules
-    def __apply_default_rules(self, ipv, transaction):
-        if ipv is "ipv4":
-            x = self.ip4tables_backend
-        elif ipv is "ipv6":
-            x = self.ip6tables_backend
+    def _select_firewall_backend(self, backend):
+        if backend != "nftables":
+            self.nftables_enabled = False
+        # even if using nftables, the other backends are enabled for use with
+        # the direct interface. nftables is used for the firewalld primitives.
+
+    def get_backend_by_name(self, name):
+        for backend in self.all_backends():
+            if backend.name == name:
+                return backend
+        raise FirewallError(errors.UNKNOWN_ERROR,
+                            "'%s' backend does not exist" % name)
+
+    def get_backend_by_ipv(self, ipv):
+        if self.nftables_enabled:
+            return self.nftables_backend
+        if ipv == "ipv4":
+            return self.ip4tables_backend
+        elif ipv == "ipv6":
+            return self.ip6tables_backend
+        elif ipv == "eb":
+            return self.ebtables_backend
+        raise FirewallError(errors.INVALID_IPV,
+                            "'%s' is not a valid backend" % ipv)
+
+    def get_direct_backend_by_ipv(self, ipv):
+        if ipv == "ipv4":
+            return self.ip4tables_backend
+        elif ipv == "ipv6":
+            return self.ip6tables_backend
+        elif ipv == "eb":
+            return self.ebtables_backend
+        raise FirewallError(errors.INVALID_IPV,
+                            "'%s' is not a valid backend" % ipv)
+
+    def is_backend_enabled(self, name):
+        if name == "ip4tables":
+            return self.ip4tables_enabled
+        elif name == "ip6tables":
+            return self.ip6tables_enabled
+        elif name == "ebtables":
+            return self.ebtables_enabled
+        elif name == "nftables":
+            return self.nftables_enabled
+        return False
+
+    def is_ipv_enabled(self, ipv):
+        if self.nftables_enabled:
+            return True
+        if ipv == "ipv4":
+            return self.ip4tables_enabled
+        elif ipv == "ipv6":
+            return self.ip6tables_enabled
+        elif ipv == "eb":
+            return self.ebtables_enabled
+        return False
+
+    def enabled_backends(self):
+        backends = []
+        if self.nftables_enabled:
+            backends.append(self.nftables_backend)
         else:
-            x = self.ebtables_backend
-        x.apply_default_rules(transaction, self._log_denied)
+            if self.ip4tables_enabled:
+                backends.append(self.ip4tables_backend)
+            if self.ip6tables_enabled:
+                backends.append(self.ip6tables_backend)
+            if self.ebtables_enabled:
+                backends.append(self.ebtables_backend)
+        return backends
+
+    def all_backends(self):
+        backends = []
+        if self.ip4tables_enabled:
+            backends.append(self.ip4tables_backend)
+        if self.ip6tables_enabled:
+            backends.append(self.ip6tables_backend)
+        if self.ebtables_enabled:
+            backends.append(self.ebtables_backend)
+        if self.nftables_enabled:
+            backends.append(self.nftables_backend)
+        return backends
+
+    def apply_default_tables(self, use_transaction=None):
+        if use_transaction is None:
+            transaction = FirewallTransaction(self)
+        else:
+            transaction = use_transaction
+
+        for backend in self.enabled_backends():
+            transaction.add_rules(backend, backend.build_default_tables())
+
+        if use_transaction is None:
+            transaction.execute(True)
 
     def apply_default_rules(self, use_transaction=None):
         if use_transaction is None:
@@ -683,18 +784,21 @@ class Firewall(object):
         else:
             transaction = use_transaction
 
-        for ipv in [ "ipv4", "ipv6", "eb" ]:
-            self.__apply_default_rules(ipv, transaction)
+        for backend in self.enabled_backends():
+            rules = backend.build_default_rules(self._log_denied)
+            transaction.add_rules(backend, rules)
 
+        ipv6_backend = self.get_backend_by_ipv("ipv6")
         if self.ipv6_rpfilter_enabled and \
-           "raw" in self.get_available_tables("ipv6"):
+           "raw" in ipv6_backend.get_available_tables():
 
             # Execute existing transaction
             transaction.execute(True)
             # Start new transaction
             transaction.clear()
 
-            self.ip6tables_backend.apply_rpfilter_rules(transaction, self._log_denied)
+            rules = ipv6_backend.build_rpfilter_rules(self._log_denied)
+            transaction.add_rules(ipv6_backend, rules)
 
             # Execute ipv6_rpfilter transaction, it might fail
             try:
@@ -718,21 +822,9 @@ class Firewall(object):
 
         log.debug1("Flushing rule set")
 
-        if self.ip4tables_enabled:
-            try:
-                self.ip4tables_backend.flush(transaction)
-            except Exception as e:
-                log.error("Failed to flush ipv4 firewall: %s" % e)
-        if self.ip6tables_enabled:
-            try:
-                self.ip6tables_backend.flush(transaction)
-            except Exception as e:
-                log.error("Failed to flush ipv6 firewall: %s" % e)
-        if self.ebtables_enabled:
-            try:
-                self.ebtables_backend.flush(transaction)
-            except Exception as e:
-                log.error("Failed to flush eb firewall: %s" % e)
+        for backend in self.all_backends():
+            rules = backend.build_flush_rules()
+            transaction.add_rules(backend, rules)
 
         if use_transaction is None:
             transaction.execute(True)
@@ -745,98 +837,50 @@ class Firewall(object):
 
         log.debug1("Setting policy to '%s'", policy)
 
-        if self.ip4tables_enabled:
-            try:
-                self.ip4tables_backend.set_policy(policy, which, transaction)
-            except Exception as e:
-                log.error("Failed to set policy of ipv4 firewall: %s" % e)
-
-        if self.ip6tables_enabled:
-            try:
-                self.ip6tables_backend.set_policy(policy, which, transaction)
-            except Exception as e:
-                log.error("Failed to set policy of ipv6 firewall: %s" % e)
-        if self.ebtables_enabled:
-            try:
-                self.ebtables_backend.set_policy(policy, which, transaction)
-            except Exception as e:
-                log.error("Failed to set policy of eb firewall: %s" % e)
+        for backend in self.enabled_backends():
+            rules = backend.build_set_policy_rules(policy, which)
+            transaction.add_rules(backend, rules)
 
         if use_transaction is None:
             transaction.execute(True)
 
     # rule function used in handle_ functions
 
-    def rule(self, ipv, rule):
-        # remove leading and trailing '"' for use with execve
-        i = 0
-        while i < len(rule):
-            x = rule[i]
-            if len(x) > 2 and x[0] == '"' and x[-1] == '"':
-                rule[i] = x[1:-1]
-            i += 1
-
-        if ipv == "ipv4":
-            # do not call if disabled
-            if self.ip4tables_enabled:
-                return self.ip4tables_backend.set_rule(rule, self._log_denied)
-        elif ipv == "ipv6":
-            # do not call if disabled
-            if self.ip6tables_enabled:
-                return self.ip6tables_backend.set_rule(rule, self._log_denied)
-        elif ipv == "eb":
-            # do not call if disabled
-            if self.ebtables_enabled:
-                return self.ebtables_backend.set_rule(rule, self._log_denied)
-        else:
+    def rule(self, backend_name, rule):
+        backend = self.get_backend_by_name(backend_name)
+        if not backend:
             raise FirewallError(errors.INVALID_IPV,
-                                "'%s' not in {'ipv4'|'ipv6'|'eb'}" % ipv)
+                                "'%s' is not a valid backend" % backend_name)
 
-        return ""
+        if not self.is_backend_enabled(backend_name):
+            return ""
 
-    def rules(self, ipv, rules):
+        return backend.set_rule(rule, self._log_denied)
+
+    def rules(self, backend_name, rules):
         _rules = rules[:]
 
-        backend = None
-        if ipv == "ipv4":
-            if self.ip4tables_enabled:
-                # do not call if disabled
-                backend = self.ip4tables_backend
-        elif ipv == "ipv6":
-            if self.ip6tables_enabled:
-                # do not call if disabled
-                backend = self.ip6tables_backend
-        elif ipv == "eb":
-            if self.ebtables_enabled:
-                # do not call if disabled
-                backend = self.ebtables_backend
-        else:
-            raise FirewallError(errors.INVALID_IPV,
-                                "'%s' not in {'ipv4'|'ipv6'|'eb'}" % ipv)
-
+        backend = self.get_backend_by_name(backend_name)
         if not backend:
+            raise FirewallError(errors.INVALID_IPV,
+                                "'%s' is not a valid backend" % backend_name)
+
+        if not self.is_backend_enabled(backend_name):
             return ""
 
         if self._individual_calls or \
            not backend.restore_command_exists or \
-           (ipv == "eb" and not self.ebtables_backend.restore_noflush_option):
+           (backend_name == "ebtables" and not self.ebtables_backend.restore_noflush_option):
             for i,rule in enumerate(_rules):
-                # remove leading and trailing '"' for use with execve
-                j = 0
-                while j < len(rule):
-                    x = rule[j]
-                    if len(x) > 2 and x[0] == '"' and x[-1] == '"':
-                        rule[j] = x[1:-1]
-                    j += 1
-
                 try:
                     backend.set_rule(rule)
                 except Exception as msg:
+                    log.debug1(traceback.format_exc())
                     log.error("Failed to apply rules. A firewall reload might solve the issue if the firewall has been modified using ip*tables or ebtables.")
                     log.error(msg)
                     for rule in reversed(_rules[:i]):
                         try:
-                            backend.set_rule(ipXtables.reverse_rule(rule))
+                            backend.set_rule(backend.reverse_rule(rule))
                         except Exception:
                             # ignore errors here
                             pass
